//----------------------------------------------------------
//-------  Parameter-less SGA		            ------------
//-------  - General definitions	     	    ------------
//----------------------------------------------------------

# ifndef SGA_H
# define SGA_H

//#include <stdio.h>
//#include <math.h>
//#include <string.h>
//#include <float.h>

//------- individual's structure definition
typedef struct {
  int *genes;		//--- binary vector
  double fitness;
} tind;

//------- Population's performance parameters observed
typedef struct {
  double fbest;   // best fitness in the population (best-so-far)
  double favg;    // average fitness of the chromosomes
  double fworst;  // fitness of the worst chromosome
  double fsum;    // sum of fitness values over the whole population
  int indbest;    // index of the best individual
  int indworst;   // index of the worst individual
} tstat;

//------- Overall statistics
typedef struct {
  double fbest;   // best-so-far
  double favg;    // average fitness of the best population
  int popSize;    // size of the best population
  int when;		 // number of fitness function evaluations needed to find the best solution
} toverallstat;

//------- SGA control parameters
extern int nTournament;	// parameter of tournament selection
extern int pCross;			// probability of crossover [%]
extern int pMutation;		// probability of mutation [%]

//------- Parameter-less GA control parameters
extern int maxGenes;			// fixed length of chromosomes
extern int *popSize;			// population sizes of all running SGAs
extern int smallestPopSize;		// current smallest population in the system
extern int maxEvaluations;		// stopping criterion - overall maximal number of fitness evaluations cummulated over all SGAs
extern int currEvaluations;		// current number of calculated fitness function evaluations
extern tind **populations;		// populations of all SGAs;
extern int nSGA;				// number of currently running SGAs
extern int maxSGA;				// maximal number of running SGAs
extern int counter;				// global generation counter
extern int counterBase;			// the base of the global generation counter (referred to as the parameter 'm' in the articles)
extern tstat *statdata;			// array of population statistics for all SGAs

//------- Others
extern int rand_seed;			// seeds for pseudo-random generator

# endif