//---------------------------------------------------------------
//---------------------------------------------------------------
//------------  Simple Genetic Algorithm             ------------
//------------  - binary representation              ------------  
//------------  - generational replacement strategy  ------------  
//---------------------------------------------------------------
//---------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <conio.h>
#include <time.h>
#include <string.h>
#include <float.h>
//----------------
#include "general.h"
#include "individual.h"
#include "generation_sga.h"

//------- SGA parameters
int nTournament;		// parameter of tournament selection
int pCross;				// probability of crossover [%]
int pMutation;			// probability of mutation [%]
int popId;
tind *newpop, *backpop, *offspring; // old and new population

//------------------------------------------------------
//------- SGA functions
//------------------------------------------------------

//------- Population statistics
void statistics(int popId){
	int i;
	double pom=0.0;

	//------- init statdata
	statdata[popId].fbest = -1.0e100;
	statdata[popId].fworst = 1.0e100;
	statdata[popId].favg = 0.0;
	statdata[popId].fsum = 0.0;
	statdata[popId].indbest = 0;
	statdata[popId].indworst = 0;
	for(i=0; i<popSize[popId]; i++){
		if(statdata[popId].fbest <= populations[popId][i].fitness){
			statdata[popId].fbest = populations[popId][i].fitness;
			statdata[popId].indbest = i;
		}
		if(statdata[popId].fworst >= populations[popId][i].fitness){
			statdata[popId].fworst = populations[popId][i].fitness;
			statdata[popId].indworst = i;
		}
		statdata[popId].fsum += populations[popId][i].fitness;
	}
	statdata[popId].favg = statdata[popId].fsum / popSize[popId];
}

//----------------------------------------------------------------------
//------- reports results
void report_results(tstat *statdata,tind *pop, int evals){
  
  printf("(%d)	fbest: %.10f 	favg: %.10f\n", evals, statdata->fbest, statdata->favg);
}

//----------------------------------------------------------------------
//------- copies the individual from pop_from to pop_to
void copy_ind(int ind_from,tind *pop_from,int ind_to,tind *pop_to){
  int i;

  for(i=0; i<maxGenes; i++){
	  pop_to[ind_to].genes[i] = pop_from[ind_from].genes[i];
  }
  pop_to[ind_to].fitness = pop_from[ind_from].fitness;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------ Tournament selection                   
//------   - tournamentSize individuals take part in tournament
int select_tournament(tind *pop, int tournamentSize){
	int cand, tempbest;

	//------- 1st competitor
	tempbest = rand()%popSize[popId]; //--- best-so-far candidate
	tournamentSize--;

	//------- try remaining competitors
	while(tournamentSize > 0){ 
		cand = rand() % popSize[popId]; 
		if(pop[cand].fitness > pop[tempbest].fitness)
			tempbest = cand;
		//-------
		tournamentSize--;
	}
	return tempbest;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//------ Chooses parents using select() strategy                   
void choose_parents(tind *pop, int *par1, int *par2){

	//------- 1st parent
	*par1 = select_tournament(pop, nTournament);

	//------- chooses 2nd parent to be different than the 1st one
	do{
		*par2 = select_tournament(pop, nTournament);
	}while(*par1 == *par2);
}

//------- Initializes population with given index popId, population size, crossover rate 
//------- and the tournament selection parameter.
//------- Returns 1 if the initialization went OK; otherwise returns 0.
int initPopulation(int popId, int *evals){
	int i,j;

	//--- setup the population size
	if(popId == 0)
		popSize[popId] = smallestPopSize;
	else
		popSize[popId] = 2 * popSize[popId-1];
	smallestPopSize = 2 * popSize[popId];	// whenever a new population is initialized, 
										// no smaller population can be initializaed any more

	//--- allocates memory for chromosomes
	if((populations[popId] = (tind *)malloc(popSize[popId] * sizeof(tind))) == NULL){
		printf("\nNot enough memory.");
		return 0;	//--- not enough memory
	}
	for(i=0; i<popSize[popId]; i++){
		if((populations[popId][i].genes = (int *)malloc(maxGenes * sizeof(int))) == NULL){
			printf("\nNot enough memory.");
			return 0;	//--- not enough memory
		}
	}

	//--- initializes chromosomes
	//--- randomly generates chromosomes with the probability of generating 1 of P(1)=0.5 
	for(i=0; i<popSize[popId]; i++){ 
		for(j=0; j<maxGenes; j++){
			if(rand()%100 < 50){              
				populations[popId][i].genes[j] = 1;
			}
			else{
				populations[popId][i].genes[j] = 0;
			}
		}
	}

	//------- evaluate chromosomes
	for(i=0; i<popSize[popId]; i++){
		get_fitness(&(populations[popId][i]));
		(*evals)++;
	}
	//------- calculate population statistics
	statistics(popId);

	printf("\n*** Population %d created:\t fitness=%.3f (%.3f)", popSize[popId], statdata[popId].favg, statdata[popId].fbest);

	return 1;	//--- everything is OK
}

//------- Destroys population with given index popId
void destroyPopulation(int popId){
	int i,j;

	//--- releases memory allocated to chromosomes of the given population
	for(i=0; i<popSize[popId]; i++)
		free(populations[popId][i].genes);
	free(populations[popId]);
	populations[popId] = NULL;
}

//------- Shifts population popId to the left and resets all population's records at the position popId
void shiftPopulationToLeft(int popId){
	int i,j;

	//------- 1. shift population popId to the left
	populations[popId-1] = populations[popId];	
	popSize[popId-1] = popSize[popId];
	statdata[popId-1] = statdata[popId];
	//------- 2. reset all records for population popId
	populations[popId] = NULL;
	popSize[popId] = 0;
}

//------- Performs one generation with the population popId
void doGeneration(int popId, int *evals){

	int par1, par2;         // indexes of parents
	int chroms_generated;   // a number of newly generated chromosomes
	int i;

	//------- allocate memory for new population and offspring
	newpop = (tind *)malloc(popSize[popId] * sizeof(tind));
	for(i=0; i<popSize[popId]; i++){
		newpop[i].genes = (int *)malloc(maxGenes * sizeof(int));
	}
	offspring = (tind *)malloc(2 * sizeof(tind));
	for(i=0; i<2; i++){
		offspring[i].genes = (int *)malloc(maxGenes * sizeof(int));
	}

	//------- elitism
	chroms_generated = 0;
	copy_ind(statdata[popId].indbest, populations[popId], chroms_generated++, newpop);

	//------- create new population of chromosomes (generational replacement strategy)
	do{
		choose_parents(populations[popId], &par1, &par2);	// choose parents
		//------- create two new chromosomes
		if(rand()%100 < pCross){
			crossover(populations[popId][par1].genes, populations[popId][par2].genes, offspring[0].genes, offspring[1].genes);
		}
		else{
			copy_ind(par1, populations[popId], 0, offspring);
			copy_ind(par2, populations[popId], 1, offspring);
		}
		//------- evaluate offspring
		get_fitness(&(offspring[0])); (*evals)++;
		get_fitness(&(offspring[1])); (*evals)++;
		//------- replacement
		copy_ind(0, offspring, chroms_generated++, newpop);
		if(chroms_generated < popSize[popId])
			copy_ind(1, offspring, chroms_generated++, newpop);
	}while(chroms_generated < popSize[popId]);

	//------- switch populations
	backpop = populations[popId];
	populations[popId] = newpop;
	newpop = backpop;
	//------- calculate population statistics
	statistics(popId);

	//------- free memory allocated for the new population and offspring
	for(i=0; i<popSize[popId]; i++){
		free(newpop[i].genes);
	}
	free(newpop);
	for(i=0; i<2; i++){
		free(offspring[i].genes);
	}
	free(offspring);

	printf("\n  - Population %d generated:\t fitness=%.3f (%.3f)", popSize[popId], statdata[popId].favg, statdata[popId].fbest);
}

//------- Checks the populations whether the relation favg(i)>=favg(i+1) for all i=1...nSGA-1.
//------- Returns the index of the largest population 'i' for which the relation favg(i)>=favg(i+1) does not hold.
//------- Otherwise, returns -1.
int checkAverageFitness(tstat *statdata){
	int i;
	int compFlag = -1;	// Contains an index of the largest population for which the favg(i)>=favg(i+1) does not hold.

	//------- If there is at most one population running in the system, then it can not be compared to anything else.
	if(nSGA <= 1)
		return compFlag;

	//------- Otherwise compare all subsequent populations.
	for(i=0; i<nSGA-1; i++)
		if(statdata[i].favg < statdata[i+1].favg)
			compFlag = i;

	return compFlag;
}

//------- Checks the convergence rate of given population
//------- Returns 1 if the population has completely converged; otherwise returns 0.
int checkConvergence(int popId){
	int i, j, k;

	for(i=0; i<popSize[popId]-1; i++){
		for(j=i+1; j<popSize[popId]; j++)
			for(k=0; k<maxGenes; k++)
				if(populations[popId][i].genes[k] != populations[popId][j].genes[k])
					return 0;	// population has NOT converged
	}

	return 1;	// population is completely homogeneous
}



